/*
 * book.c - C source for GNU CHESS
 *
 * Revision: 1991-04-15
 *
 * Copyright (C) 1986, 1987, 1988, 1989, 1990 Free Software Foundation, Inc.
 * Copyright (c) 1988, 1989, 1990  John Stanback
 *
 * This file is part of CHESS.
 *
 * CHESS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY.  No author or distributor accepts responsibility to anyone for
 * the consequences of using it or for whether it serves any particular
 * purpose or works at all, unless he says so in writing.  Refer to the CHESS
 * General Public License for full details.
 *
 * Everyone is granted permission to copy, modify and redistribute CHESS, but
 * only under the conditions described in the CHESS General Public License. A
 * copy of this license is supposed to have been given to you along with
 * CHESS so you can know your rights and responsibilities.  It should be in a
 * file named COPYING.  Among other things, the copyright notice and this
 * notice must be preserved on all copies.
 */
#include "schah.h"
extern char mvstr[4][6];
short bookcount = 0;
static struct bookentry
{
  unsigned long bookkey;
  unsigned long bookbd;
  unsigned short move;
  unsigned short hint;
  unsigned char count;
  unsigned char flags;
} OpenBook[BOOKSIZE];
static struct bookentry *BookTable[BKTBLSIZE];
void
GetOpenings (void)

/*
 * Read in the Opening Book file and parse the algebraic notation for a move
 * into an unsigned integer format indicating the from and to square. Create
 * a linked list of opening lines of play, with entry->next pointing to the
 * next line and entry->move pointing to a chunk of memory containing the
 * moves. More Opening lines of up to 100 half moves may be added to
 * schah.book.
 */
#ifndef BOOK
#define BOOK "/usr/games/lib/schah.book"
#endif /* BOOK */
{
  FILE *fd;
  register struct bookentry *OB, *OC;
  register short int i, f, t;
  char opening[80];
  char msg[80];
  short int xside, doit, c, side;
  short int rf, rt;
  unsigned short mv;


  if ((fd = fopen (BOOK, "r")) == NULL)
    fd = fopen ("schah.book", "r");
  if (fd != NULL)
    {
      OC = NULL;
      /* setvbuf(fd,buffr,_IOFBF,2048); */
      Book = NULL;
      side = white;
      xside = black;
      for (OB = OpenBook; OB < &OpenBook[BOOKSIZE]; OB++)
	OB->count = 0;
      for (i = 0; i < BKTBLSIZE; i++)
	{
	  BookTable[i] = &OpenBook[BOOKSIZE / BKTBLSIZE * i];
	}
      hashbd = hashkey = 0;
      for (i = 0; i < 64; i++)
	{
	  board[i] = Stboard[i];
	  color[i] = Stcolor[i];
	}
      i = 0;

      while ((c = parse (fd, &mv, side, opening)) >= 0)
	if (c == 1)
	  {

	    /*
	     * if not first move of an opening and first time we have
	     * seen it save next move as hint
	     */
	    if (i && OB->count == 1)
	      OB->hint = mv & 0x7fff;
	    OC = OB;		/* save for end marking */
	    doit = true;

	    /*
	     * see if this position and move already exist from some
	     * other opening
	     */
	    /* is this ethical, to offer the bad move as a hint????? */
	    for (OB = BookTable[hashkey & BOOKMASK]; OB->count; OB++)
	      {
		if (OB->bookkey == hashkey
		    && OB->bookbd == hashbd
		    && (OB->flags & SIDEMASK) == side
		    && OB->move == mv)
		  {

		    /*
		     * yes so just bump count - count is used to choose
		     * opening move in proportion to its presence in the
		     * book
		     */
		    doit = false;
		    OB->count++;
		    break;
		  }
		/* Book is hashed into BKTBLSIZE chunks based on hashkey */
		if (OB == &OpenBook[BOOKSIZE - 1])
		  OB = OpenBook;
	      }
	    /* doesn`t exist so add it to the book */
	    if (doit)
	      {
		bookcount++;
		OB->bookkey = hashkey;
		OB->bookbd = hashbd;
		OB->move = mv;
		OB->hint = 0;
		OB->count = 1;
		OB->flags = side;
	      }
	    /* now update the board and hash values */
	    /* should really check the moves as we do this, but??? */
	    f = mv >> 8 & 0x7F;
	    t = mv & 0xFF;
	    if (board[t] != no_piece)
	      {
		if (color[t] != xside)
		  {
		    algbr (f, t, false);
		    sprintf (msg, CP[211], i + 1, mvstr, opening);
		    ShowMessage (msg);
		  }
		UpdateHashbd (xside, board[t], -1, t);
	      }
	    if (board[f] == no_piece || color[f] != side)
	      {
		algbr (f, t, false);
		sprintf (msg, CP[211], i + 1, mvstr, opening);
		ShowMessage (msg);
	      }
	    UpdateHashbd (side, board[f], f, t);
	    board[t] = board[f];
	    color[t] = color[f];
	    color[f] = neutral;
	    board[f] = no_piece;
	    if (board[t] == king && (mv == BLACKCASTLE) || (mv == WHITECASTLE) || (mv == LONGBLACKCASTLE) || (mv == LONGWHITECASTLE))
	      {

		if (t > f)
		  {
		    rf = f + 3;
		    rt = t - 1;
		  }
		else
		  {
		    rf = f - 4;
		    rt = t + 1;
		  }
		board[rt] = rook;
		color[rt] = side;
		board[rf] = no_piece;
		color[rf] = neutral;
		UpdateHashbd (side, rook, rf, rt);
	      }
	    i++;
	    xside = side;
	    side = otherside[side];
	  }
	else if (c == 0 && i > 0)
	  {
	    /* Mark last move as end of an opening */
	    /* might want to terminate? */
	    OB->flags |= BOOKEND;
	    if (i > 1)
	      OC->flags |= BOOKEND;
	    /* reset for next opening */
	    side = white;
	    hashbd = hashkey = 0;
	    for (i = 0; i < 64; i++)
	      {
		board[i] = Stboard[i];
		color[i] = Stcolor[i];
	      }
	    i = 0;

	  }
      fclose (fd);
#if !defined CHESSTOOL && !defined XBOARD
      sprintf (msg, CP[213], bookcount, BOOKSIZE);
      ShowMessage (msg);
#endif
      /* set every thing back to start game */
      Book = BOOKFAIL;
      for (i = 0; i < 64; i++)
	{
	  board[i] = Stboard[i];
	  color[i] = Stcolor[i];
	}
      BKBook = Book;
    }
  else
    ShowMessage (CP[212]);
}


int
OpeningBook (unsigned short *hint, short int side)

/*
 * Go thru each of the opening lines of play and check for a match with the
 * current game listing. If a match occurs, generate a random number. If this
 * number is the largest generated so far then the next move in this line
 * becomes the current "candidate". After all lines are checked, the
 * candidate move is put at the top of the Tree[] array and will be played by
 * the program. Note that the program does not handle book transpositions.
 */

{
  short pnt;
  unsigned short m;
  unsigned r, cnt, tcnt, ccnt;
  register struct bookentry *OB, *OC;

  srand ((unsigned int) time ((long *) 0));
  m = 0;
  /* pick a number */
  r = urand () % 1000;
  cnt = 0;
  tcnt = 0;
  ccnt = 0;
  OC = NULL;

  /*
   * find all the moves for this position  - count them and get their total
   * count
   */
  for (OB = BookTable[hashkey & BOOKMASK]; OB->count; OB++)
    {
      if (OB->bookkey == hashkey
	  && OB->bookbd == hashbd
	  && ((OB->flags) & SIDEMASK) == side
	  && !(OB->move & 0x8000))
	{
	  OC = OB;
	  cnt++;
	  tcnt += OB->count;
	}
    }
  /* if only one just do it */
  if (cnt == 1)
    {
      m = OC->move;
    }
  else
    /* more than one must choose one at random */
  if (cnt > 1)

    {
      for (OC = BookTable[hashkey & BOOKMASK]; OC->count; OC++)
	{
	  if (OC->bookkey == hashkey
	      && OC->bookbd == hashbd
	      && ((OC->flags) & SIDEMASK) == side
	      && !(OC->move & 0x8000))
	    {
	      ccnt += OC->count;
	      if (((ccnt * 1000) / tcnt) >= r)
		{
		  m = OC->move;
		  break;
		}
	    }
	}
    }
  else
    {
      /* none decrement count of no finds */
      Book--;
      return false;
    }
  /* make sure the move is in the MoveList */
  for (pnt = TrPnt[1]; pnt < TrPnt[2]; pnt++)
    {
      if (((Tree[pnt].f << 8) | Tree[pnt].t) == m)
	{
	  Tree[pnt].score = 0;
	  break;
	}
    }
  /* Make sure its the best */

  pick (TrPnt[1], TrPnt[2] - 1);
  if (Tree[TrPnt[1]].score)
    {
      /* no! */
      Book--;
      return false;
    }
  /* ok pick up the hint and go */
  *hint = OC->hint;
  Book = ((OC->flags & BOOKEND) && ((urand () % 1000) > 500))?0:BOOKFAIL;
  return true;
}
